/* -*- Mode: C++; tab-width: 2; indent-tabs-mode: nil; c-basic-offset: 2 -*- */
/*
 * Pan - A Newsreader for Gtk+
 * Copyright (C) 2002-2006  Charles Kerr <charles@rebelbase.com>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; version 2 of the License.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include <cstdarg>
#include <gobject/gvaluecollector.h>
#include "pan-tree.h"

#define IS_SORTED(tree) \
  (tree->sort_column_id != GTK_TREE_SORTABLE_UNSORTED_SORT_COLUMN_ID)

/***
****
***/
struct CellSetterImpl: public PanTreeStore::CellSetter
{
  private:

    PanTreeStore * tree;
    GtkTreeIter * iter;
    PanTreeStore::Node * node;

    GValue * get_gvalue (int column)
    {
      g_assert (0<= column && column < tree->n_columns);
      return &node->columns[column];
    }

  public:

    CellSetterImpl (PanTreeStore * t, GtkTreeIter * i)
    {
      g_assert (t);
      g_assert (i);
      g_assert (IS_PAN_TREE_STORE(t));
      g_assert (i->stamp == t->stamp);

      tree = t;
      iter = i;
      node = static_cast<PanTreeStore::Node*>(iter->user_data);
    }

    virtual ~CellSetterImpl () {}

  public:

    virtual bool set_cell_gvalue (int column, const GValue& gvalue)
    {
      g_return_val_if_fail (0<=column && column<tree->n_columns, false);

      // if value isn't of the right type, transform it.
      bool converted (false);
      GValue real_value = {0, };
      const GType in_type (G_VALUE_TYPE (&gvalue));
      const GType dest_type ((*tree->column_types)[column]);
      if (!g_type_is_a (in_type, dest_type)) {
        if (!(g_value_type_compatible (in_type, dest_type) &&
              g_value_type_compatible (dest_type, in_type))) {
          g_warning ("%s: Unable to convert from %s to %s\n",
                     G_STRLOC, g_type_name (in_type), g_type_name (dest_type));
          return false;
        }
        if (!g_value_transform (&gvalue, &real_value)) {
          g_warning ("%s: Unable to make conversion from %s to %s\n",
                     G_STRLOC, g_type_name (in_type), g_type_name (dest_type));
          g_value_unset (&real_value);
          return false;
        }
        converted = true;
      }

      // set the value
      if (!converted)
        g_value_copy (&gvalue, &node->columns[column]);
      else {
        g_value_copy (&real_value, &node->columns[column]);
        g_value_unset (&real_value);
      }
      return true;
    }

#define SET_VALUE(COLUMN,VALUE,TYPE,FUNC) \
    GValue * gval = get_gvalue (COLUMN); \
    if (G_VALUE_TYPE(gval) == TYPE) { \
      FUNC (gval, VALUE); \
      return true; \
    } else { \
      GValue tmp = { 0, }; \
      g_value_init (&tmp, TYPE); \
      FUNC (&tmp, VALUE); \
      const bool ok = set_cell_gvalue (column, tmp); \
      g_value_unset (&tmp); \
      return ok; \
    }
    bool set_cell_int (int column, int value) {
      SET_VALUE (column, value, G_TYPE_INT, g_value_set_int);
    }
    bool set_cell_uint (int column, unsigned int value) {
      SET_VALUE (column, value, G_TYPE_UINT, g_value_set_uint);
    }
    bool set_cell_string (int column, const char * value) {
      SET_VALUE (column, value, G_TYPE_STRING, g_value_set_string);
    }
    bool give_cell_string (int column, char * value) {
      SET_VALUE (column, value, G_TYPE_STRING, g_value_take_string);
    }
    bool set_cell_pointer (int column, void* value) {
      SET_VALUE (column, value, G_TYPE_POINTER, g_value_set_pointer);
    }
    bool set_cell_long (int column, long value) {
      SET_VALUE (column, value, G_TYPE_LONG, g_value_set_long);
    }
    bool set_cell_ulong (int column, unsigned long value) {
      SET_VALUE (column, value, G_TYPE_ULONG, g_value_set_ulong);
    }
    bool set_cell_char (int column, char value) {
      SET_VALUE (column, value, G_TYPE_CHAR, g_value_set_char);
    }
    bool set_cell_bool (int column, bool value) {
      SET_VALUE (column, value, G_TYPE_BOOLEAN, g_value_set_boolean);
    }
};

/***
****
***/

namespace
{
  PanTreeStore::Node* get_node (PanTreeStore * tree, GtkTreeIter * iter)
  {
    g_assert (tree);
    g_assert (iter);
    g_assert (iter->stamp == tree->stamp);
    PanTreeStore::Node * n = static_cast<PanTreeStore::Node*>(iter->user_data);
    g_assert (n);
    return n;
  }
}

namespace
{
  void set_iter (GtkTreeIter   * iter,
                 int             stamp,
                 PanTreeStore::Node * node)
  {
    iter->stamp = stamp;
    iter->user_data = node;
    iter->user_data2 = 0;
    iter->user_data3 = 0;
  }

  void invalidate_iter (GtkTreeIter * iter)
  {
    set_iter (iter, 0, 0);
  }

  GtkTreeModelFlags get_flags (GtkTreeModel *tree_model)
  {
    return GTK_TREE_MODEL_ITERS_PERSIST;
  }

  gint get_n_columns (GtkTreeModel *tree)
  {
    return PAN_TREE_STORE(tree)->n_columns;
  }

  GType get_column_type (GtkTreeModel * tree,
                            gint           n)
  {
    return (*((PanTreeStore*)(tree))->column_types)[n];
  }

  gboolean get_iter (GtkTreeModel *tree_model,
                     GtkTreeIter  *iter,
                     GtkTreePath  *path)
  {
    PanTreeStore * tree = PAN_TREE_STORE(tree_model);
    g_return_val_if_fail (tree, false);
    g_return_val_if_fail (path, false);

    // find the article that corresponds to this path.
    PanTreeStore::Node * node = tree->root;
    const int depth = gtk_tree_path_get_depth (path);
    const int * indices = gtk_tree_path_get_indices (path);
    for (int i=0; i<depth; ++i) {
      const int index = *indices++;
      if (index<0 || index>=node->n_children)
        return false;
      node = node->children[index];
    }

    set_iter (iter, tree->stamp, node);
    return true;
  }

  GtkTreePath* get_path (GtkTreeModel *tree_model,
                         GtkTreeIter  *iter)
  {
    PanTreeStore * tree = PAN_TREE_STORE(tree_model);
    g_return_val_if_fail (tree, false);
    g_return_val_if_fail (iter, false);
    g_return_val_if_fail (iter->stamp == tree->stamp, false);

    GtkTreePath * path = gtk_tree_path_new ();
    PanTreeStore::Node * node = get_node (tree, iter);
    while (node && node!=tree->root) {
      gtk_tree_path_prepend_index (path, node->child_index);
      node = node->parent;
    }
    return path;
  }

  void get_value (GtkTreeModel * tree_model,
                  GtkTreeIter  * iter,
                  gint           column,
                  GValue       * dest_value)
  {
    g_assert (iter);
    g_assert (dest_value);
    PanTreeStore * tree = PAN_TREE_STORE(tree_model);
    g_assert (tree);
    g_assert (iter->stamp == tree->stamp);
    g_assert (0<=column && column<tree->n_columns);

    PanTreeStore::Node * node = get_node (tree, iter);
    g_value_init (dest_value, (*tree->column_types)[column]);
    g_value_copy (&node->columns[column], dest_value);
  }

  gboolean iter_next (GtkTreeModel * model,
                      GtkTreeIter  * iter)
  {
    PanTreeStore * tree = PAN_TREE_STORE(model);
    g_return_val_if_fail (tree, false);
    g_return_val_if_fail (iter, false);
    g_return_val_if_fail (iter->stamp == tree->stamp, false);

    PanTreeStore::Node * node = get_node (tree, iter);
    const int new_index = node->child_index + 1;
    if (new_index >= node->parent->n_children) {
      invalidate_iter (iter);
      return false;
    }

    set_iter (iter, tree->stamp, node->parent->children[new_index]);
    return true;
  }

  gboolean iter_children (GtkTreeModel * model,
                          GtkTreeIter  * iter,
                          GtkTreeIter  * parent)
  {
    PanTreeStore * tree = PAN_TREE_STORE(model);
    g_return_val_if_fail (tree, false);
    g_return_val_if_fail (iter, false);
    g_return_val_if_fail (!parent || (parent->stamp == tree->stamp), false);

    PanTreeStore::Node * node = parent ? get_node (tree, parent) : tree->root;
    if (node->children.empty()) {
      invalidate_iter (iter);
      return false;
    }

    set_iter (iter, tree->stamp, node->children[0]);
    return true;
  }

  gint iter_n_children (GtkTreeModel * model,
                        GtkTreeIter  * iter)
  {
    PanTreeStore * tree = PAN_TREE_STORE(model);
    g_return_val_if_fail (tree, 0);
    g_return_val_if_fail (!iter || iter->stamp == tree->stamp, 0);

    const PanTreeStore::Node * node = iter ? get_node(tree, iter) : tree->root;
    return node->n_children;
  }

  gboolean iter_has_child (GtkTreeModel *model,
                           GtkTreeIter  *iter)
  {
    return iter_n_children (model, iter) != 0;
  }

  gboolean iter_nth_child (GtkTreeModel * model,
                           GtkTreeIter  * iter,
                           GtkTreeIter  * parent,
                           gint           n)
  {
    PanTreeStore * tree = PAN_TREE_STORE(model);
    g_return_val_if_fail (tree, false);
    g_return_val_if_fail (iter, false);
    g_return_val_if_fail (!parent || parent->stamp == tree->stamp, false);

    PanTreeStore::Node * node = parent ? get_node (tree, parent) : tree->root;
    if (n<0 || n>=node->n_children) {
      invalidate_iter (iter);
      return false;
    }

    set_iter (iter, tree->stamp, node->children[n]);
    return true;
  }

  gboolean iter_parent (GtkTreeModel * model,
                        GtkTreeIter  * iter,
                        GtkTreeIter  * child)
  {
    PanTreeStore * tree = PAN_TREE_STORE(model);
    g_return_val_if_fail (tree, false);
    g_return_val_if_fail (child, false);
    g_return_val_if_fail (iter, false);
    g_return_val_if_fail (child->stamp == tree->stamp, false);

    PanTreeStore::Node * node = get_node (tree, child);
    const bool has_parent = node->parent != tree->root;
    if (has_parent)
      set_iter (iter, tree->stamp, node->parent);
    else
      invalidate_iter (iter);
    return has_parent;
  }
}

namespace
{
  gboolean get_sort_column_id (GtkTreeSortable  * sortable,
                               gint             * sort_column_id,
                               GtkSortType      * order)
  {
    PanTreeStore * tree (PAN_TREE_STORE (sortable));
    g_return_val_if_fail (tree, false);

    if (sort_column_id)
       *sort_column_id = tree->sort_column_id;
    if (order)
       *order = tree->order;

    return (tree->sort_column_id != GTK_TREE_SORTABLE_DEFAULT_SORT_COLUMN_ID)
        && (tree->sort_column_id != GTK_TREE_SORTABLE_UNSORTED_SORT_COLUMN_ID);
  }

  gboolean has_sort_func (GtkTreeSortable *sortable, int col)
  {
    PanTreeStore * tree (PAN_TREE_STORE (sortable));
    g_return_val_if_fail (tree, false);
    std::map<int,PanTreeStore::SortInfo>::const_iterator it =
      tree->sort_info->find (col);
    return (it!=tree->sort_info->end()) && (it->second.sort_func!=0);
  }

  struct SortTuple
  {
    int offset;
    PanTreeStore::Node * node;
    GtkTreeIter iter;
  };

  struct SortHelperData
  {
    GtkTreeModel * model;
    PanTreeStore::SortInfo& sort_info;
    GtkSortType order;
    SortHelperData (GtkTreeModel           * m,
                    PanTreeStore::SortInfo & s,
                    GtkSortType              o):
      model(m), sort_info(s), order(o) {}
  };

  int tuple_compare_func (gconstpointer a_gpointer,
                          gconstpointer b_gpointer,
                          gpointer user_data)
  {
    const SortTuple * a = (const SortTuple*) a_gpointer;
    const SortTuple * b = (const SortTuple*) b_gpointer;
    const SortHelperData * help = (const SortHelperData*) user_data;
    const int val = (help->sort_info.sort_func) (
                       help->model,
                       const_cast<GtkTreeIter*>(&a->iter),
                       const_cast<GtkTreeIter*>(&b->iter),
                       help->sort_info.user_data);
    return help->order==GTK_SORT_ASCENDING ? val : -val;
  }

  void pan_tree_store_sort_helper (PanTreeStore            * tree,
                                   PanTreeStore::SortInfo  & sort_info,
                                   PanTreeStore::Node      * parent,
                                   bool                      recurse)
  {
    const int n = parent->n_children;
    if (n < 2) // no need to sort [0...1] children
      return;

    // build a temporary array to sort
    std::vector<SortTuple> sort_array (n);
    for (int i=0; i<n; ++i) {
      SortTuple& tuple = sort_array[i];
      tuple.offset = i;
      tuple.node = parent->children[i];
      set_iter (&tuple.iter, tree->stamp, tuple.node);
    }

    // sort 'em...
    SortHelperData help (GTK_TREE_MODEL(tree), sort_info, tree->order);
    g_qsort_with_data (&sort_array.front(),
                       sort_array.size(),
                       sizeof(SortTuple),
                       tuple_compare_func,
                       &help);

    // update the child indices...
    for (int i=0; i<n; ++i) {
      PanTreeStore::Node * child = sort_array[i].node;
      child->child_index = i;
      parent->children[i] = child;
      g_assert (child->parent == parent);
    }

    // let the world know we've changed
#if GLIB_CHECK_VERSION(2,10,0)
    int * new_order = (int*) g_slice_alloc (n * sizeof(int));
#else
    int * new_order = g_new (gint, n);
#endif
    bool differ = false;
    for (int i=0; i<n; ++i) {
      differ |= (i != sort_array[i].offset);
      new_order[i] = sort_array[i].offset;
    }

    // tell the world...
    GtkTreeIter tmp_iter, *parent_iter;
    GtkTreePath * path;
    if (parent == tree->root) {
      parent_iter = 0;
      path = gtk_tree_path_new ();
    } else {
      set_iter (&tmp_iter, tree->stamp, parent);
      parent_iter = &tmp_iter;
      path = gtk_tree_model_get_path (help.model, parent_iter);
    }
    gtk_tree_model_rows_reordered (help.model, path, parent_iter, new_order);
    gtk_tree_path_free (path);
#if GLIB_CHECK_VERSION(2,10,0)
    g_slice_free1 (sizeof(int)*n, new_order);
#else
    g_free (new_order);
#endif

    for (int i=0; recurse && i<n; ++i) {
      PanTreeStore::Node * child = parent->children[i];
      pan_tree_store_sort_helper (tree, sort_info, child, recurse);
    }
  }

  void pan_tree_store_sort (PanTreeStore * tree)
  {
    g_assert (IS_PAN_TREE_STORE(tree));

    if (tree->sort_column_id != GTK_TREE_SORTABLE_UNSORTED_SORT_COLUMN_ID)
    {
      g_assert (has_sort_func (GTK_TREE_SORTABLE(tree), tree->sort_column_id));
      PanTreeStore::SortInfo& info = (*tree->sort_info)[tree->sort_column_id];
      pan_tree_store_sort_helper (tree, info, tree->root, true);
    }
  }

  void pan_tree_store_sort_iter_changed (PanTreeStore      * tree,
                                         GtkTreeIter  * iter,
                                         gint           column)
  {
    // FIXME
  }


  void set_sort_column_id (GtkTreeSortable * sortable,
                           gint              sort_column_id,
                           GtkSortType       order)
  {
    PanTreeStore * tree (PAN_TREE_STORE (sortable));
    g_return_if_fail (tree);

    // if no change, there's nothing to do...
    if (tree->sort_column_id==sort_column_id && tree->order==order)
      return;

    // sanity checks...
    if (sort_column_id != GTK_TREE_SORTABLE_UNSORTED_SORT_COLUMN_ID) {
      g_return_if_fail (tree->sort_info->count(sort_column_id) != 0);
      g_return_if_fail (tree->sort_info->find(sort_column_id)->second.sort_func != 0);
    }

    tree->sort_column_id = sort_column_id;
    tree->order = order;
    gtk_tree_sortable_sort_column_changed (sortable);
    pan_tree_store_sort (tree);
  }

  gboolean has_default_sort_func (GtkTreeSortable *sortable)
  {
    return has_sort_func (sortable, GTK_TREE_SORTABLE_DEFAULT_SORT_COLUMN_ID);
  }

  void set_sort_func (GtkTreeSortable        *sortable,
                      gint                    col,
                      GtkTreeIterCompareFunc  func,
                      gpointer                data,
                      GtkDestroyNotify        destroy)
  {
    PanTreeStore * tree (PAN_TREE_STORE (sortable));
    g_return_if_fail (tree);

    (*tree->sort_info)[col].assign (func, data, destroy);

    if (tree->sort_column_id == col)
      pan_tree_store_sort (tree);
  }

  void set_default_sort_func (GtkTreeSortable        * s,
                              GtkTreeIterCompareFunc   f,
                              gpointer                 p,
                              GtkDestroyNotify         d)
  {
    set_sort_func (s, GTK_TREE_SORTABLE_DEFAULT_SORT_COLUMN_ID, f, p, d);
  }
}

namespace
{
  GObjectClass *parent_class (0);

  typedef std::vector<GValue> values_t;

  void pan_tree_node_unset_values (PanTreeStore::Node * node, int n_columns)
  {
    if (node->columns)
    {
      GValue * it = node->columns;
      for (int i=0; i<n_columns; ++i, ++it)
        g_value_unset (it);
#if GLIB_CHECK_VERSION(2,10,0)
      g_slice_free1 (sizeof(GValue)*n_columns, node->columns);
#else
      g_free (node->columns);
#endif
      node->columns = 0;
    }
  }

  void
  pan_tree_finalize (GObject *object)
  {
    PanTreeStore * tree = (PanTreeStore*)object;
    const int n_columns = tree->n_columns;
    typedef std::deque<PanTreeStore::Node>::iterator it_t;
    for (it_t it(tree->nodes->begin()), end(tree->nodes->end()); it!=end; ++it)
      pan_tree_node_unset_values (&*it, n_columns);
   
    delete tree->column_types;
    delete tree->reuse_me;
    delete tree->nodes;
    delete tree->sort_info;
    g_free (tree->columns_template);

    (*parent_class->finalize) (object);
  }

  void
  pan_tree_class_init (PanTreeStoreClass *klass)
  {
    GObjectClass *object_class;

    parent_class = (GObjectClass*) g_type_class_peek_parent (klass);
    object_class = (GObjectClass*) klass;

    object_class->finalize = pan_tree_finalize;
  }

  void
  pan_tree_model_init (GtkTreeModelIface *iface)
  {
    iface->get_flags       = get_flags;
    iface->get_n_columns   = get_n_columns;
    iface->get_column_type = get_column_type;
    iface->get_iter        = get_iter;
    iface->get_path        = get_path;
    iface->get_value       = get_value;
    iface->iter_next       = iter_next;
    iface->iter_children   = iter_children;
    iface->iter_has_child  = iter_has_child;
    iface->iter_n_children = iter_n_children;
    iface->iter_nth_child  = iter_nth_child;
    iface->iter_parent     = iter_parent;
  }

  void
  pan_tree_sortable_init (GtkTreeSortableIface *iface)
  {
    iface->get_sort_column_id = get_sort_column_id;
    iface->set_sort_column_id = set_sort_column_id;
    iface->set_sort_func = set_sort_func;
    iface->set_default_sort_func = set_default_sort_func;
    iface->has_default_sort_func = has_default_sort_func;
  }

  void
  pan_tree_init (PanTreeStore * tree)
  {
    tree->stamp = g_random_int();
    tree->n_columns = 0;
    tree->nodes = new PanTreeStore::nodes_t (1);
    tree->reuse_me = new std::deque<PanTreeStore::Node*>();
    tree->root = &(*tree->nodes)[0];
    tree->column_types = new std::vector<GType>();
    tree->sort_info = new std::map<int,PanTreeStore::SortInfo>();
    tree->sort_column_id = GTK_TREE_SORTABLE_UNSORTED_SORT_COLUMN_ID;
    tree->order = GTK_SORT_ASCENDING;
  }

  GValue*
  alloc_columns (const PanTreeStore * tree)
  {
    const int n_columns = tree->n_columns;
#if GLIB_CHECK_VERSION(2,10,0)
    GValue * columns = (GValue*) g_slice_alloc0 (sizeof(GValue) * n_columns);
#else
    GValue * columns = g_new0 (GValue, n_columns);
#endif
    memcpy (columns, tree->columns_template, sizeof(GValue)*n_columns);
    return columns;
  }

  void
  pan_tree_alloc_nodes (PanTreeStore                     * tree,
                        int                                n_rows,
                        std::vector<PanTreeStore::Node*> & setme)
  {
    setme.resize (n_rows);
    PanTreeStore::Node ** sit = &setme.front();

    PanTreeStore::nodes_t& nodes (*tree->nodes);
    const int old_size = nodes.size ();
    nodes.resize (old_size + n_rows);
    for (PanTreeStore::nodes_t::iterator i(nodes.begin()+old_size), end(nodes.end()); i!=end; ++i) {
      PanTreeStore::Node *node = *sit++ = &*i;
      node->columns = alloc_columns (tree);
    }
  }
}

/***
****
***/

GType
pan_tree_store_get_type (void)
{
  static GType pan_tree_type (0);
  if (pan_tree_type)
    return pan_tree_type;

  static const GTypeInfo pan_tree_info = {
    sizeof (PanTreeStoreClass),
    NULL, // base_init
    NULL, // base_finalize
    (GClassInitFunc) pan_tree_class_init,
    NULL, // class finalize
    NULL, // class_data
    sizeof (PanTreeStore),
    0, // n_preallocs
    (GInstanceInitFunc) pan_tree_init
  };

  pan_tree_type = g_type_register_static (
    G_TYPE_OBJECT, "PanTreeStore", &pan_tree_info, (GTypeFlags)0);
  static const GInterfaceInfo tree_model_info =
    { (GInterfaceInitFunc)pan_tree_model_init, 0, 0 };
  g_type_add_interface_static (
    pan_tree_type, GTK_TYPE_TREE_MODEL, &tree_model_info);
  static const GInterfaceInfo sortable_info =
    { (GInterfaceInitFunc)pan_tree_sortable_init, 0, 0 };
  g_type_add_interface_static (
    pan_tree_type, GTK_TYPE_TREE_SORTABLE, &sortable_info);
  return pan_tree_type;
}

PanTreeStore *
pan_tree_store_new (int n_columns, ...)
{
  g_return_val_if_fail (n_columns>0, 0);
  PanTreeStore* tree = (PanTreeStore*) g_object_new (PAN_TREE_STORE_TYPE, NULL);

  va_list args;
  va_start (args, n_columns);
  for (int i=0; i<n_columns; ++i) {
    const GType type = va_arg (args, GType);
    tree->column_types->push_back (type);
  }
  va_end (args);
  tree->n_columns = tree->column_types->size();

  const int n = tree->n_columns;
  tree->columns_template = g_new0 (GValue, n);
  for (int i=0; i<n; ++i) {
    GValue& value = tree->columns_template[i];
    g_value_init (&value, (*tree->column_types)[i]);
  }

  return tree;
}

/***
****
***/

void
pan_tree_store_insert_n  (PanTreeStore               * tree,
                          GtkTreeIter                * iter,
                          GtkTreeIter                * parent_it,
                          PanTreeStore::RowSetter    & row_setter,
                          int                          n_rows,
                          int                          position)
{
  g_return_if_fail (IS_PAN_TREE_STORE(tree));
  g_return_if_fail (position >= 0);
  g_return_if_fail (n_rows > 0);

  // insert the rows
  PanTreeStore::Node * parent_node =
    parent_it ? get_node (tree, parent_it) : tree->root;
  std::vector<PanTreeStore::Node*> new_children;
  pan_tree_alloc_nodes (tree, n_rows, new_children);
  const int old_size = parent_node->n_children;
  position = std::min (position, old_size);
  parent_node->children.insert (parent_node->children.begin()+position,
                                new_children.begin(),
                                new_children.end());
  parent_node->n_children += n_rows;
  PanTreeStore::Node** it = &parent_node->children[position];
  for (int i=position, new_size=old_size+n_rows; i<new_size; ++i) {
    PanTreeStore::Node * node (*it++);
    node->child_index = i;
    node->parent = parent_node;
  }

  // set the return iter
  GtkTreeIter walk;
  set_iter (&walk, tree->stamp, parent_node->children[position]);
  if (iter)
    *iter = walk;

  // emit some signals
  GtkTreeModel * model = GTK_TREE_MODEL(tree);
  GtkTreePath * path = get_path (model, &walk);
  it = &parent_node->children[position];
  for (int i=0; i<n_rows; ++i) {
    PanTreeStore::Node * node (*it++);
    set_iter (&walk, tree->stamp, node);
    CellSetterImpl csi (tree, &walk);
    row_setter.set_row (i, csi);
    gtk_tree_model_row_inserted (model, path, &walk);
    gtk_tree_path_next (path);
  }
  if (!old_size && parent_node!=tree->root) {
    gtk_tree_path_up (path);
    gtk_tree_model_row_has_child_toggled (model, path, parent_it);
  }
  gtk_tree_path_free (path);
}

void
pan_tree_store_append_n  (PanTreeStore  * tree,
                          GtkTreeIter   * iter,
                          GtkTreeIter   * parent,
                          PanTreeStore::RowSetter     & row_setter,
                          int             n_rows)
{
  pan_tree_store_insert_n (tree, iter, parent, row_setter, n_rows, INT_MAX);
}

/***
****
***/

namespace 
{
  struct DoNothingRowSetter: public PanTreeStore::RowSetter {
    virtual ~DoNothingRowSetter() {}
    virtual void set_row (int n, PanTreeStore::CellSetter&)  {}
  };
}

void
pan_tree_store_insert (PanTreeStore * tree,
                       GtkTreeIter  * iter,
                       GtkTreeIter  * parent,
                       gint           pos)
{
  DoNothingRowSetter row_setter;
  pan_tree_store_insert_n (tree, iter, parent, row_setter, 1, pos);
}

void
pan_tree_store_insert_before (PanTreeStore * tree,
                              GtkTreeIter  * iter,
                              GtkTreeIter  * parent,
                              GtkTreeIter  * sibling)
{
  const int pos = sibling ? get_node(tree,sibling)->child_index : INT_MAX;
  pan_tree_store_insert (tree, iter, parent, pos);
}

void
pan_tree_store_insert_after (PanTreeStore * tree,
                             GtkTreeIter  * iter,
                             GtkTreeIter  * parent,
                             GtkTreeIter  * sibling)
{
  const int pos = sibling ? get_node(tree,sibling)->child_index+1 : 0;
  pan_tree_store_insert (tree, iter, parent, pos);
}

void pan_tree_store_prepend  (PanTreeStore * tree,
                              GtkTreeIter  * iter,
                              GtkTreeIter  * parent)
{
  pan_tree_store_insert (tree, iter, parent, 0);
}

void
pan_tree_store_append (PanTreeStore * tree,
                       GtkTreeIter * iter,
                       GtkTreeIter * parent)
{
  pan_tree_store_insert (tree, iter, parent, INT_MAX);
}

/***
****
***/

void
pan_tree_store_set (PanTreeStore * tree,
                    GtkTreeIter  *iter,
                    ...)
{
  va_list var_args;
  va_start (var_args, iter);
  pan_tree_store_set_valist (tree, iter, var_args);
  va_end (var_args);
}

void
pan_tree_store_set_valist (PanTreeStore * tree,
                           GtkTreeIter  * iter,
                           va_list        var_args)
{
  g_return_if_fail (IS_PAN_TREE_STORE(tree));
  g_return_if_fail (iter);
  g_return_if_fail (iter->stamp == tree->stamp);

  if (IS_SORTED(tree))
    g_return_if_fail (has_sort_func (GTK_TREE_SORTABLE(tree), tree->sort_column_id));

  bool maybe_need_sort (false); //func != pan_tree_data_list_compare_func;
  bool emit_signal (false);
  int column = va_arg (var_args, gint);
  while (column != -1)
  {
    if (column >= tree->n_columns) {
      g_warning ("%s: Invalid column number %d added to iter (remember to end your list of columns with a -1)", G_STRLOC, column);
      break;
    }

    GValue value = { 0, };
    g_value_init (&value, (*tree->column_types)[column]);
    gchar * error = 0;
    G_VALUE_COLLECT (&value, var_args, 0, &error);
    if (error) {
      g_warning ("%s: %s", G_STRLOC, error);
      g_free (error);
      /* we purposely leak the value here, it might not be
       * in a sane state if an error condition occured */
      break;
    }

    CellSetterImpl csi (tree, iter);
    const bool did_set = csi.set_cell_gvalue (column, value);
    emit_signal |= did_set;
    g_value_unset (&value);

    maybe_need_sort |= column == tree->sort_column_id;

    column = va_arg (var_args, gint);
  }

  if (maybe_need_sort && IS_SORTED(tree))
    pan_tree_store_sort_iter_changed (tree, iter, tree->sort_column_id);

  if (emit_signal) {
    GtkTreePath * path = gtk_tree_model_get_path (GTK_TREE_MODEL(tree), iter);
    gtk_tree_model_row_changed (GTK_TREE_MODEL(tree), path, iter);
    gtk_tree_path_free (path);
  }
}

/***
****
***/
namespace
{
  struct NodeWalker
  {
    virtual ~NodeWalker() {}
    virtual void operator()(PanTreeStore::Node* node) = 0;
  };

  void node_walk (NodeWalker& func, PanTreeStore::Node * node)
  {
    for (int i=0, n=node->n_children; i<n; ++i)
      node_walk (func, node->children[i]);
    func (node);
  }

  struct UnsetNodeWalker: public NodeWalker
  {
    PanTreeStore * tree;
    UnsetNodeWalker(PanTreeStore* t): tree(t) {}
    virtual ~UnsetNodeWalker() {}
    virtual void operator()(PanTreeStore::Node* node) {
      pan_tree_node_unset_values (node, tree->n_columns);
      node->parent = 0;
      node->child_index = -1;
      //tree->reuse_me->push_back (node);
    }
  };
}

gboolean
pan_tree_store_remove (PanTreeStore *tree, GtkTreeIter *iter)
{
  g_return_val_if_fail (IS_PAN_TREE_STORE(tree), false);
  g_return_val_if_fail (iter, false);
  g_return_val_if_fail (iter->stamp == tree->stamp, false);

  GtkTreeModel * model = GTK_TREE_MODEL (tree);
  GtkTreePath * path = gtk_tree_model_get_path (model, iter);

  // check the pre-removal state
  PanTreeStore::Node * node = get_node (tree, iter);
  PanTreeStore::Node * parent = node->parent;

  // unthread the node
  const int pos = node->child_index;
  parent->children.erase (parent->children.begin() + pos);
  --parent->n_children;
  if (pos < parent->n_children) {
    PanTreeStore::Node ** it (&parent->children[pos]);
    for (int i=pos; i<parent->n_children; ++i, ++it)
      (*it)->child_index = i;
  }

  // revalidate the iterator
  const bool iter_is_valid_now (pos < parent->n_children);
  if (iter_is_valid_now)
    set_iter (iter, tree->stamp, parent->children[pos]);
  else
    invalidate_iter (iter);

  // let the world know the row was removed
  gtk_tree_model_row_deleted (model, path);

  // if the parent's has-children state has changed, let the world know that too
  if (parent!=tree->root && !parent->n_children) {
    GtkTreeIter pit;
    set_iter (&pit, tree->stamp, parent);
    gtk_tree_path_up (path);
    gtk_tree_model_row_has_child_toggled (model, path, &pit);
  }

  gtk_tree_path_free (path);

  // add `node' and all its descendants back to the reuse-me pool
  UnsetNodeWalker walker (tree);
  node_walk (walker, node);

  return iter_is_valid_now;
}

gpointer 
pan_tree_store_peek_value (PanTreeStore * tree, GtkTreeIter * iter, int column)
{
  g_assert (tree);
  g_assert (iter);
  g_assert (tree->stamp == iter->stamp);
  g_assert (0<=column && column<tree->n_columns);

  PanTreeStore::Node * node = get_node (tree, iter);
  return g_value_peek_pointer (&node->columns[column]);
}
